#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/7/24 11:55:53

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2
import numpy as np
from PIL import Image
#===============================================================================
# skimage 图像格式        通道：RGB 像素值：[0.0,1.0]      (h,w)
# Opencv  图像格式        通道：BGR 像素值：[0,255]        (h,w)
# PIL     图像格式        (w,h)
#===============================================================================

__all__ = ['is_pil_image', 'is_numpy_image', 'gray2bgr', 'gray2rgb', 'bgr2gray', 'rgb2gray',
           'rgb2bgr', 'bgr2rgb', 'cv2pil', 'pil2cv', 'cv2sk', 'sk2cv']

def is_pil_image(img):
    return isinstance(img, Image.Image)

def is_numpy_image(img):
    return isinstance(img, np.ndarray) and img.ndim in {2, 3}

def gray2bgr(img):
    #cvtColor(src, code[, dst[, dstCn]]) -> dst
    if img.ndim == 3 and img.shape[-1] == 3:
        return img
    elif img.ndim == 2:
        return cv2.cvtColor(img, cv2.COLOR_GRAY2BGR)
    return img


def gray2rgb(img):
    if img.ndim == 3 and img.shape[-1] == 3:
        return img
    elif img.ndim == 2:
        return cv2.cvtColor(img, cv2.COLOR_GRAY2RGB)
    return img


def bgr2gray(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            return cv2.cvtColor(img, cv2.COLOR_BGR2GRAY )
        elif img.shape[-1] == 4:
            return cv2.cvtColor(img, cv2.COLOR_BGRA2GRAY )
    elif img.ndim == 2:
        return img
    return img


def rgb2gray(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            return cv2.cvtColor(img, cv2.COLOR_RGB2GRAY )
        elif img.shape[-1] == 4:
            return cv2.cvtColor(img, cv2.COLOR_RGBA2GRAY )
    elif img.ndim == 2:
        return img
    return img


def rgb2bgr(img):
    return cv2.cvtColor(img, cv2.COLOR_RGB2BGR )


def bgr2rgb(img):
    return cv2.cvtColor(img, cv2.COLOR_BGR2RGB )


def cv2pil(img):
    if img.ndim == 3 and img.shape[-1] == 3:
        image = Image.fromarray(cv2.cvtColor(img,cv2.COLOR_BGR2RGB), mode = 'RGB')
    else:
        image = Image.fromarray(img, mode = 'L')
    return image


def pil2cv(img):
    if img.mode == 'RGB':
        image = cv2.cvtColor(np.asarray(img), cv2.COLOR_RGB2BGR)
    elif img.mode == 'RGBA':
        image = cv2.cvtColor(np.asarray(img), cv2.COLOR_RGBA2BGRA)
    elif img.mode == 'L':
        image = np.asarray(img)
    return image


def cv2sk(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            img = cv2.cvtColor(img,cv2.COLOR_BGR2RGB)
        elif img.shape[-1] == 4:
            img = cv2.cvtColor(img,cv2.COLOR_BGRA2RGBA)
    return img


def sk2cv(img):
    if img.ndim == 3:
        if img.shape[-1] == 3:
            img = cv2.cvtColor(img,cv2.COLOR_RGB2BGR)
        elif img.shape[-1] == 4:
            img = cv2.cvtColor(img,cv2.COLOR_RGBA2BGRA)
    return img


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/27 11:21:00

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import itertools, re, time
from functools import partial
import torch
import torchvision.models as tmodels
import torchvision.transforms as T
import PIL
from laok.core.kconfig import path_model as _path_model
import os.path
from laok.core.ktime import Timer
#===============================================================================
# 模型后缀一般是 .pt .pth
#===============================================================================
__all__ = ['save_model', 'load_model', 'save_model_params', 'load_model_params',
           'get_model_list', 'get_model_file', 'get_model']
############################################################### 模型加载和保存
def save_model(model, filename):
    '''该方式无需自定义网络模型，保存时已把网络结构保存
    :param model:   net model instance
    :param filename: model filename
    :return:
    '''
    torch.save(model, filename)

def load_model(filename):
    '''加载模型
    :param filename: 模型文件
    :return: 模型
    '''
    return torch.load(filename)

def save_model_params(model, filename):
    ''' 保存模型的参数
    :param model: 模型
    :param filename: 模型文件
    :return: None
    '''
    torch.save(model.state_dict(), filename)

def _state_dict_name_normal(state_dict):
    # '.'s are no longer allowed in module names, but previous _DenseLayer
    # has keys 'norm.1', 'relu.1', 'conv.1', 'norm.2', 'relu.2', 'conv.2'.
    # They are also in the checkpoints in model_urls. This pattern is used
    # to find such keys.
    pattern = re.compile(
        r'^(.*denselayer\d+\.(?:norm|relu|conv))\.((?:[12])\.(?:weight|bias|running_mean|running_var))$')
    for key in list(state_dict.keys()):
        res = pattern.match(key)
        if res:
            new_key = res.group(1) + res.group(2)
            state_dict[new_key] = state_dict[key]
            del state_dict[key]

def load_model_params(model, filename, replace_dot_name = False):
    '''加载模型参数
    :param model: 模型
    :param filename: 文件名字
    :param replace_dot_name: 替换模型层中.包含在名字中的层
    :return: 模型
    '''
    _obj = torch.load(filename)
    state_dict = None
    if isinstance(_obj, torch.nn.Module):
        state_dict = _obj.state_dict()
    elif isinstance(_obj, dict):
        state_dict = _obj

    if replace_dot_name and state_dict:
        _state_dict_name_normal(state_dict)
    # print('load ', filename)
    model.load_state_dict(state_dict)
    return model

############################################################### 模型仓库
# 默认缓冲在 C:\Users\user\.cache\torch\hub\checkpoints
_classify_models = {
    'alexnet': { 'creator': tmodels.alexnet, 'path' : 'alexnet-owt-4df8aa71.pth'},
    'densenet121':  { 'creator': tmodels.densenet121,'path' : 'densenet121-a639ec97.pth'},
    'densenet169': { 'creator': tmodels.densenet169,'path' : 'densenet169-b2777c0a.pth'},
    'densenet201': { 'creator': tmodels.densenet201,'path' : 'densenet201-c1103571.pth'},
    'densenet161': { 'creator': tmodels.densenet161,'path' : 'densenet161-8d451a50.pth'},
    'googlenet': { 'creator': partial(tmodels.googlenet,init_weights=False),'path' : 'googlenet-1378be20.pth'},
    'inception_v3_google': { 'creator': partial(tmodels.inception_v3,init_weights=False),'path' : 'inception_v3_google-1a9a5a14.pth'},
    'mnasnet0_5': { 'creator': tmodels.mnasnet0_5,'path' : 'mnasnet0.5_top1_67.823-3ffadce67e.pth'},
    'mnasnet0_75': { 'creator': tmodels.mnasnet0_75,'path' : ''},
    'mnasnet1_0': { 'creator': tmodels.mnasnet1_0,'path' : 'mnasnet1.0_top1_73.512-f206786ef8.pth'},
    'mnasnet1_3': { 'creator': tmodels.mnasnet1_3,'path' : ''},
    'mobilenet_v2': { 'creator': tmodels.mobilenet_v2,'path' : 'mobilenet_v2-b0353104.pth'},
    'mobilenet_v3_small': { 'creator': tmodels.mobilenet_v3_small,'path' : 'mobilenet_v3_small-047dcff4.pth'},
    'mobilenet_v3_large': { 'creator': tmodels.mobilenet_v3_large,'path' : 'mobilenet_v3_large-8738ca79.pth'},
    'resnet18': { 'creator': tmodels.resnet18,'path' : 'resnet18-5c106cde.pth'},
    'resnet34': { 'creator': tmodels.resnet34,'path' : 'resnet34-333f7ec4.pth'},
    'resnet50': { 'creator': tmodels.resnet50,'path' : 'resnet50-19c8e357.pth'},
    'resnet101': { 'creator': tmodels.resnet101,'path' : 'resnet101-5d3b4d8f.pth'},
    'resnet152': { 'creator': tmodels.resnet152,'path' : 'resnet152-b121ed2d.pth'},
    'resnext50_32x4d': { 'creator': tmodels.resnext50_32x4d,'path' : 'resnext50_32x4d-7cdf4587.pth'},
    'resnext101_32x8d': { 'creator': tmodels.resnext101_32x8d,'path' : 'resnext101_32x8d-8ba56ff5.pth'},
    'wide_resnet50_2': { 'creator': tmodels.wide_resnet50_2,'path' : 'wide_resnet50_2-95faca4d.pth'},
    'wide_resnet101_2': { 'creator': tmodels.wide_resnet101_2,'path' : 'wide_resnet101_2-32ee1156.pth'},
    'shufflenet_v2_x0_5': { 'creator': tmodels.shufflenet_v2_x0_5,'path' : 'shufflenetv2_x0.5-f707e7126e.pth'},
    'shufflenet_v2_x1_0': { 'creator': tmodels.shufflenet_v2_x1_0,'path' : 'shufflenetv2_x1-5666bf0f80.pth'},
    'shufflenet_v2_x1_5': { 'creator': tmodels.shufflenet_v2_x1_5,'path' : ''},
    'shufflenet_v2_x2_0': { 'creator': tmodels.shufflenet_v2_x2_0,'path' : ''},
    'squeezenet1_0': { 'creator': tmodels.squeezenet1_0,'path' : 'squeezenet1_0-a815701f.pth'},
    'squeezenet1_1': { 'creator': tmodels.squeezenet1_1,'path' : 'squeezenet1_1-f364aa15.pth'},
    'vgg11': { 'creator': tmodels.vgg11,'path' : 'vgg11-bbd30ac9.pth'},
    'vgg13': { 'creator': tmodels.vgg13,'path' : 'vgg13-c768596a.pth'},
    'vgg16': { 'creator': tmodels.vgg16,'path' : 'vgg16-397923af.pth'},
    'vgg19': { 'creator': tmodels.vgg19,'path' : 'vgg19-dcbb9e9d.pth'},
    'vgg11_bn': { 'creator': tmodels.vgg11_bn,'path' : 'vgg11_bn-6002323d.pth'},
    'vgg13_bn': { 'creator': tmodels.vgg13_bn,'path' : 'vgg13_bn-abd245e5.pth'},
    'vgg16_bn': { 'creator': tmodels.vgg16_bn,'path' : 'vgg16_bn-6c64b313.pth'},
    'vgg19_bn': { 'creator': tmodels.vgg19_bn,'path' : 'vgg19_bn-c79401a0.pth'},
}

_detection_models = {
    'fasterrcnn_mobilenet_v3_large_320_fpn': {'creator': tmodels.alexnet, 'path': 'fasterrcnn_mobilenet_v3_large_320_fpn-907ea3f9.pth'},
    'fasterrcnn_mobilenet_v3_large_fpn': {'creator': tmodels.alexnet, 'path': 'fasterrcnn_mobilenet_v3_large_fpn-fb6a3cc7'},
    'fasterrcnn_resnet50_fpn': {'creator': tmodels.alexnet, 'path': 'fasterrcnn_resnet50_fpn_coco-258fb6c6.pth'},
    'keypointrcnn_resnet50_fpn': {'creator': tmodels.alexnet, 'path': 'keypointrcnn_resnet50_fpn_coco-fc266e95.pth'},
    'maskrcnn_resnet50_fpn_coco': {'creator': tmodels.alexnet, 'path': 'maskrcnn_resnet50_fpn_coco-bf2d0c1e.pth'},
    'retinanet_resnet50_fpn_coco': {'creator': tmodels.alexnet, 'path': 'retinanet_resnet50_fpn_coco-eeacb38b.pth'},
}


# def _segmentation_load_weights(model, arch_type, backbone, progress):
#     arch = arch_type + '_' + backbone + '_coco'
#     model_url = model_urls.get(arch, None)
#     if model_url is None:
#         raise NotImplementedError('pretrained {} is not supported as of now'.format(arch))
#     else:
#         state_dict = load_state_dict_from_url(model_url, progress=progress)
#         model.load_state_dict(state_dict)
#
# tmodels.segmentation._load_weights = _segmentation_load_weights

_segmentation_models = {
    'deeplabv3_mobilenet_v3_large': {'creator': tmodels.segmentation.deeplabv3_mobilenet_v3_large, 'path': 'deeplabv3_mobilenet_v3_large-fc3c493d.pth'},
    'deeplabv3_resnet50': {'creator': tmodels.segmentation.deeplabv3_resnet50, 'path': 'deeplabv3_resnet50_coco-cd0a2569.pth'},
    'deeplabv3_resnet101': {'creator': tmodels.segmentation.deeplabv3_resnet101, 'path': 'deeplabv3_resnet101_coco-586e9e4e.pth'},
    'fcn_resnet50': {'creator': tmodels.segmentation.fcn_resnet50, 'path': 'fcn_resnet50_coco-1167a1af.pth'},
    'fcn_resnet101': {'creator': tmodels.segmentation.fcn_resnet101, 'path': 'fcn_resnet101_coco-7ecb50ca.pth'},
    'lraspp_mobilenet_v3_large': {'creator': tmodels.segmentation.lraspp_mobilenet_v3_large, 'path': 'lraspp_mobilenet_v3_large-d234d4ea.pth'},
}

_video_models = {
    'mc3_18': {'creator': tmodels.alexnet, 'path': 'mc3_18-a90a0ba3.pth'},
    'r2plus1d_18': {'creator': tmodels.alexnet, 'path': 'r2plus1d_18-91a641e6.pth'},
    'r3d_18': {'creator': tmodels.alexnet, 'path': 'r3d_18-b3b3357e.pth'},
}

_quantization_models = {
    'googlenet_fbgemm': {'creator': tmodels.alexnet, 'path': 'googlenet_fbgemm-c00238cf.pth'},
    'inception_v3_google_fbgemm': {'creator': tmodels.alexnet, 'path': 'inception_v3_google_fbgemm-71447a44.pth'},
    'mobilenet_v2_qnnpack': {'creator': tmodels.alexnet, 'path': 'mobilenet_v2_qnnpack_37f702c5.pth'},
    'resnet18_fbgemm': {'creator': tmodels.alexnet, 'path': 'resnet18_fbgemm_16fa66dd.pth'},
    'resnet50_fbgemm': {'creator': tmodels.alexnet, 'path': 'resnet50_fbgemm_bf931d71.pth'},
    'resnext101_32x8_fbgemm': {'creator': tmodels.alexnet, 'path': 'resnext101_32x8_fbgemm_09835ccf.pth'},
    'shufflenetv2_x1_fbgemm': {'creator': tmodels.alexnet, 'path': 'shufflenetv2_x1_fbgemm-db332c57.pth'},
}

def get_model_list(model_type = None):
    '''
    :param model_type: 'classify','detection','segmentation', 'video', 'quantization',None
    :return:
    '''
    model_type_list = []

    if model_type is not None:
        model_type = model_type.lower()

    if model_type is None:
        model_type_list.extend([_classify_models, _detection_models, _segmentation_models, _video_models, _quantization_models])
    elif model_type == 'classify':
        model_type_list.append(_classify_models)
    elif model_type == 'detection':
        model_type_list.append(_detection_models)
    elif model_type == 'segmentation':
        model_type_list.append(_segmentation_models)
    elif model_type == 'video':
        model_type_list.append(_video_models)
    elif model_type == 'quantization':
        model_type_list.append(_quantization_models)
    for k in itertools.chain(*model_type_list):
        yield k

def get_model_file(model_name):
    fpath = None
    model_name = model_name.lower()
    if model_name in _classify_models:
        fpath = _path_model('Torch/classify/' + _classify_models[model_name]['path'])
    elif model_name in _detection_models:
        fpath = _path_model('Torch/detection/' + _detection_models[model_name]['path'])
    elif model_name in _segmentation_models:
        fpath = _path_model('Torch/segmentation/' + _segmentation_models[model_name]['path'])
    elif model_name in _video_models:
        fpath = _path_model('Torch/video/' + _video_models[model_name]['path'])
    elif model_name in _quantization_models:
        fpath = _path_model('Torch/quantization/' + _quantization_models[model_name]['path'])
    if fpath and os.path.isfile(fpath):
        return fpath

def get_model(model_name, load_params = True, **kws):
    ctor = None
    model_name = model_name.lower()
    if model_name in _classify_models:
        ctor = _classify_models[model_name]['creator']
    elif model_name in _detection_models:
        ctor = _detection_models[model_name]['creator']
    elif model_name in _segmentation_models:
        ctor = _segmentation_models[model_name]['creator']
    elif model_name in _video_models:
        ctor = _video_models[model_name]['creator']
    elif model_name in _quantization_models:
        ctor = _quantization_models[model_name]['creator']

    if ctor is None:
        return

    model = ctor()
    if load_params:
        model_file = get_model_file(model_name)
        if model_file:
            replace_dot_name = 'densenet' in model_name
            load_model_params(model, model_file, replace_dot_name= replace_dot_name)
    return model

############################################################### 模型训练和测试
# model.train() ：启用 BatchNormalization 和 Dropout
# model.eval() ：不启用 BatchNormalization 和 Dropout


def get_device():
    return torch.device("cuda:0" if _g_has_cuda else "cpu")

_g_has_cuda = torch.cuda.is_available()
_g_device = get_device()

def transform_image_eval():
    return T.Compose([T.Resize(256),
                      T.CenterCrop(224),
                      T.ToTensor(),
                      T.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])])

def transform_image_train():
    return T.Compose([
        T.RandomResizedCrop(size=256, scale=(0.8, 1.0)),
        T.RandomRotation(degrees=15),
        T.RandomHorizontalFlip(),
        T.CenterCrop(size=224),
        T.ToTensor(),
        T.Normalize([0.485, 0.456, 0.406],
                    [0.229, 0.224, 0.225])])

def inference_image_classify(img, model, classes = None, topk=1, transform = None, device = None):
    if isinstance(img, str):
        _img = PIL.Image.open(img).convert('RGB')
        if transform is None:
            transform = transform_image_eval()
    else:
        _img = img

    if transform == 'default':
        transform = transform_image_eval()

    img_t = transform(_img) if transform else _img
    batch_t =  torch.unsqueeze(img_t, 0) if img_t.dim() == 3 else img_t

    # Validation - No gradient tracking needed
    with torch.no_grad():
        model.eval()

        if _g_has_cuda and device is None:
            model = model.to(_g_device)
            batch_t = batch_t.to(_g_device)

        t1 = time.time()
        out = model(batch_t)
        t2 = time.time()

    percentage = torch.nn.functional.softmax(out, dim=1)[0] * 100
    values, indices = torch.sort(out, descending=True)

    ret = []
    for i in range(topk):
        idx = indices[0][i].item()
        val = values[0][i].item()
        score = percentage[idx].item()
        label = classes[idx] if classes and idx < len(classes) else ''
        obj = {
            'index':idx,
            'value': val ,
            'label': label,
            'score': score,
            'inference_time':  f"{1000* (t2-t1):.6}(ms)"
        }

        if isinstance(img, str):
            obj['file'] = os.path.basename(img)
        ret.append(obj)
    return ret

def model_train():
    pass

def model_validate():
    pass

def model_inference():
    pass

def model_inference_image_classify():
    pass


def train(data_loader, model, optimizer = None, criterion = None, epoches = 10, batches_display=100, device = None):
    '''训练模型的基本步骤
    :param data_loader:
    :param model:
    :param optimizer:
    :param criterion:
    :param epoches:
    :param batches_display:
    :param device:
    :return:
    '''
    if criterion is None:
        criterion = torch.nn.CrossEntropyLoss()

    if optimizer is None:
        optimizer = torch.optim.SGD(model.parameters(), lr=0.001, momentum=0.9)

    if _g_has_cuda and device is None:
        print('=====Convert to [cuda] device')
        model.to(_g_device)

    if device:
        model.to(_g_device)

    print('======Trainning start....')
    totalTimer = Timer()
    batchTimer = Timer()
    for epoch in range(epoches):  # loop over the dataset multiple times
        running_loss = 0.0

        sampleCount = 0
        batchTimer.restart()
        for i, data in enumerate(data_loader, 0):
            # get the inputs; data is a list of [inputs, labels]
            inputs, targets = data
            sampleCount += inputs.size(0)

            if _g_has_cuda and device is None:
                inputs, targets = inputs.to(_g_device), targets.to(_g_device)

            if device:
                inputs, targets = inputs.to(_g_device), targets.to(_g_device)

            # zero the parameter gradients
            optimizer.zero_grad()

            # forward + backward + optimize
            outputs = model(inputs)
            loss = criterion(outputs, targets)

            loss.backward()
            optimizer.step()

            # print statistics
            running_loss += loss.item()
            if i % batches_display == (batches_display-1):  # print every 2000 mini-batches
                print('[Ep:%d,%d] loss: %.6f time:%.6f(s)' %
                      (epoch + 1, i + 1, running_loss/sampleCount, batchTimer.elapse()))
                running_loss = 0.0
                sampleCount = 0
                batchTimer.restart()

    print('======Trainning finished, cosume time:%.6f(s)....' % totalTimer.elapse())

def eval_classify(data_loader, model, classes, topk=1, device = None):
    class_correct = list(0. for i in classes)
    class_total = list(0. for i in classes)

    if _g_has_cuda and device is None:
        model.to(_g_device)
    model.eval()

    for data in data_loader:
        inputs, targets = data

        if _g_has_cuda and device is None:
            inputs, targets = inputs.to(_g_device), targets.to(_g_device)

        outputs = model(inputs)
        _, predicted = torch.max(outputs, topk)
        c = (predicted == targets).squeeze()

        for i,target in enumerate(targets):
            class_correct[target] += c[i].item()
            class_total[target] += 1

    correct_total = 0
    total = 0.0
    for i,cls in enumerate(classes):
        correct_total += class_correct[i]
        total += class_total[i]
        if class_total[i]:
            percent = 100 * class_correct[i] / class_total[i]
        else:
            percent = 0
        print('Accuracy of %5s : %2d %%' % (cls, percent))
    print('Final Accuracy : %2d %%' % (correct_total/total * 100) )


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/17 21:11:08

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import time as _time
import sys
from contextlib import contextmanager as _context
from functools import wraps as _wraps
#===============================================================================
# 
#===============================================================================

__all__ = ['auto_timer', 'Timer']

@_context
def auto_timer(msg = '', need_print = True):
    t1 = _time.time()
    yield
    t2 = _time.time()
    if need_print:
        print(f'{msg} [time:{(t2-t1)*1000}(ms)]')

def deco_time(stream = None):
    '''
    装饰器,记录时间
    '''
    def _w1(func):
        @_wraps(func)
        def _w2(*args, **kwargs):
            start = _time.time()
            res = func(*args, **kwargs)
            stop = _time.time()
            if _time is None:
                stream = sys.stdout
            stream.write(f"[{ func.__name__ }] use time [{ 1000*(stop-start) }(ms)]\n")
            return res
        return _w2
    return _w1


class Timer:
    def __init__(self):
        self.restart()

    def restart(self):
        self._t1 = _time.time()

    def elapse(self):
        '''
        time in seconds since the Epoch
        :return:
        '''
        return _time.time() - self._t1




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/4/9 09:09:27

@author: laok
@copyright: Apache License, Version 2.0
'''
from torch.utils.data import DataLoader
import torchvision.datasets as datasets

# ===============================================================================
# 
# ===============================================================================
_vision_datasets = {
    'ImageFolder': { 'creator': datasets.ImageFolder },
    'MNIST': { 'creator': datasets.MNIST },
}


def _find_dataset(name):
    name = name.lower()
    for k,v in _vision_datasets.items():
        if k.lower() == name:
            return v


def get_dataset(name, **kws):
    _dataset = _find_dataset(name)
    if _dataset :
        return _dataset['creator'](**kws)


def get_dataset_loader(name):
    pass





#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/27 14:18:09

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import os.path as opath
#===============================================================================
# 
#===============================================================================

def path_unix(fpath):
    return fpath.replace('\\', '/')

def path_win(fpath):
    return fpath.replace('/', '\\')

def env_file(key, filename):
    for pth in os.getenv('path').split(';'):
        full_path = opath.join(pth, filename)
        if opath.exists(full_path):
            return path_unix(full_path)

def exe_path_file(filename):
    if opath.exists(filename):
        return path_unix(filename)
    return env_file('path', filename)

def env_files(key, filename):
    for pth in os.getenv(key).split(';'):
        full_path = opath.join(pth, filename)
        if opath.exists(full_path):
            yield path_unix(full_path)

def exe_path_files(filename):
    if opath.exists(filename):
        yield path_unix(filename)
    for full_path in env_files('path', filename):
        yield path_unix(full_path)

def path_join(*args):
    return path_unix(os.path.join(*args))

def path_abs(fpath):
    return path_unix(os.path.abspath(fpath))

def path_exist(fpath):
    return fpath and os.path.exists(fpath)

def path_is_file(fpath):
    return fpath and os.path.isfile(fpath)

def path_is_dir(fpath):
    return fpath and os.path.isdir(fpath)

def path_ensure_parent(fpath):
    parent, _ = opath.split(fpath)
    os.makedirs(parent, exist_ok=True)

def path_parent(fpath):
    '''fpath: 输入路径名字
       ret: 返回父路径
    '''
    return path_unix(opath.split(fpath)[0])

def path_filename(fpath):
    '''file_path: 输入路径名字
       ret: 返回文件名字
    '''
    if isinstance(fpath, str):
        return opath.split(fpath)[1]
    if isinstance(fpath, list):
        return [opath.split(f)[1] for f in fpath]

def path_basename(fpath):
    '''file_path: 输入路径名字
       ret: 返回去掉后缀的文件名字
    '''
    fname = opath.split(fpath)[1]
    return opath.splitext(fname)[0]

def path_ext(fpath, need_dot = True):
    '''file_path: 输入路径名字
       ret: 返回后缀
    '''
    fname = opath.split(fpath)[1]
    ext = opath.splitext(fname)[1]
    if ext and not need_dot:
        return ext[1:]
    return ext

def path_replace_ext(file_path, ext):
    '''file_path: 输入路径名字
       ext: 需要替换的后缀名字
       ret: 返回替换后的路径
    '''
    if isinstance(ext, str) and not ext.startswith('.'):
        ext = '.' + ext
    new_file = opath.splitext(file_path)[0] + ext
    return path_unix(new_file)

def path_replace_basename(file_path, basename):
    '''
        file_path：输入路径名字
        basename: 替换的文件名字,保留后缀
        ret: 返回替换后的路径
    '''
    _dir, _fname = opath.split(file_path)
    _basename, _ext = opath.splitext(_fname)
    return path_unix(opath.join(_dir, basename + _ext))

def path_replace_filename(file_path, filename):
    '''
        file_path：输入路径名字
        filename: 替换的文件名字
        ret: 返回替换后的路径
    '''
    _dir, _fname = opath.split(file_path)
    return path_unix(opath.join(_dir, filename))

def path_replace_parent(file_path, parent):
    '''
        file_path：输入路径名字
        parent: 替换的父路径
        ret: 返回替换后的路径
    '''
    _dir, _fname = opath.split(file_path)
    return path_unix(opath.join(parent, _fname))

def _ensure_list(suffix_list):
    if suffix_list is None:
        return []
    if isinstance(suffix_list, list):
        return suffix_list
    if isinstance(suffix_list, str):
        return suffix_list.split(';')
    return suffix_list

def _file_name_endswith(fname, suffix_list):
    return suffix_list and any( (fname.endswith(s) for s in suffix_list) )

def files_under(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [.txt,.py];用于筛选文件
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fdir_name, _sub_fdirs, files in os.walk(dir_name):
        for fname in files:
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(os.path.join(fdir_name, fname) if need_join_dir else fname)

def files_current(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [.txt,.py];用于筛选文件
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fname in os.listdir(dir_name):
        fpath = os.path.join(dir_name, fname)
        if os.path.isfile(fpath):
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(fpath if need_join_dir else fname)


def dirs_under(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [.txt,.py]
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fdir_name, sub_fdirs, files in os.walk(dir_name):
        for fname in sub_fdirs:
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(os.path.join(fdir_name, fname) if need_join_dir else fname)

def dirs_current(dir_name, suffix_list = None, need_join_dir=True):
    '''
    dir_name: 路径
    suffix_list: 后缀列表, 比如: [test, test2];用于筛选路径
    ret: 返回生成器
    '''
    suffix_list = _ensure_list(suffix_list)
    for fname in os.listdir(dir_name):
        fpath = os.path.join(dir_name, fname)
        if os.path.isdir(fpath):
            if not suffix_list or _file_name_endswith(fname, suffix_list):
                yield path_unix(fpath if need_join_dir else fname)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/6/4 14:41:10

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from .dump_description import dump_description_help
from .lktest_runner import lktest_run
from pprint import pprint
import sys as _sys
#===============================================================================
# 
#===============================================================================

def print_eval(st):
    _frm = _sys._getframe(1)
    print( "{} [:] {}".format(st, eval(st, _frm.f_globals, _frm.f_locals)) )

def print_type(t):
    import inspect
    type_name = None
    if inspect.ismodule(t):
        type_name = 'ismodule'
    if inspect.isclass(t):
        type_name = "isclass"
    if inspect.ismethod(t):
        type_name = "ismethod"
    if inspect.isfunction(t):
        type_name = "isfunction"
    if inspect.isgeneratorfunction(t):
        type_name = "isgeneratorfunction"
    if inspect.isgenerator(t):
        type_name = "isgenerator"
    if inspect.iscoroutinefunction(t):
        type_name = "iscoroutinefunction"
    if inspect.iscoroutine(t):
        type_name = "iscoroutine"
    if inspect.isawaitable(t):
        type_name = "isawaitable"
    if inspect.isasyncgenfunction(t):
        type_name = "isasyncgenfunction"
    if inspect.isasyncgen(t):
        type_name = "isasyncgen"
    if inspect.istraceback(t):
        type_name = "istraceback"
    if inspect.isframe(t):
        type_name = "isframe"
    if inspect.iscode(t):
        type_name = "iscode"
    if inspect.isbuiltin(t):
        type_name = "isbuiltin"
    if inspect.isroutine(t):
        type_name = "isroutine"
    if inspect.isabstract(t):
        type_name = "isabstract"
    if inspect.ismethoddescriptor(t):
        type_name = "ismethoddescriptor"
    if inspect.isdatadescriptor(t):
        type_name = "isdatadescriptor"
    if inspect.isgetsetdescriptor(t):
        type_name = "isgetsetdescriptor"
    if inspect.ismemberdescriptor(t):
        type_name = "ismemberdescriptor"
    if type_name is None:
        type_name = f"{type(t)}"
    print(type_name)


def simple_description(obj):
    for i in dir(obj):
        if not i.startswith('_'):
            attr = getattr(obj,i)
            print(f"{i}={attr}")



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/15 19:48:37

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import sys, contextlib
#===============================================================================
# 
#===============================================================================
__all__ = ['StdoutRedirect', 'stdout_redirect']


class _NullOutput:
    def __init__(self):
        pass

    def write(self, *args, **kws):
        pass

    def close(self):
        pass


class StdoutRedirect:
    def __init__(self, obj=None):
        '''
        :param obj:
            None, 表示空对象
            str, 表示文件名字
            否则, 其他类文件对象
        '''
        if obj is None:
            self._f = _NullOutput()
        elif isinstance(obj, str):
            self._f = open(obj, 'w')
        else:
            self._f = obj

    def __enter__(self):
        return self.doEnter()

    def __exit__(self, exc_type, exc_val, exc_tb):
        return self.doExit()

    def doEnter(self):
        self._old = sys.stdout
        sys.stdout = self._f
        return self

    def doExit(self):
        sys.stdout = self._old
        if hasattr(self._f, 'close'):
            self._f.close()

@contextlib.contextmanager
def stdout_redirect(filename = None):
    with StdoutRedirect(filename):
        yield




#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/17 23:09:07

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from tensorboardX import SummaryWriter
#===============================================================================
# pip install tensorboardX
# 启动服务器: tensorboard --logdir=<log_dir>
# 默认使用 http://localhost:6006/ 打开查看
#===============================================================================
DEFAULT_PATH = 'D:/AllInOne/DataRaw/TensorBoardX'
DEFAULT_PATH = r'D:\AllInOne\klibpy\algs\AREL\data\save\tensorboard\default'
def run_server(logdir=None, port=6006, host='localhost', reload_interval=5, bind_all=False, start_webbrowser=True):
    import sys
    if logdir is None:
        logdir = DEFAULT_PATH
    args = [f'--logdir={logdir}',
            f'--reload_interval={reload_interval}',
            f'--port={port}',
            f'--window_title={logdir}',
            ]
    if bind_all:
        args.append('--bind_all')
    else:
        args.append(f'--host={host}')

    sys.argv.extend(args)
    print(f'cmd={sys.argv}')

    if start_webbrowser:
        import webbrowser
        webbrowser.open(url=f'http://localhost:{port}')

    from tensorboard.main import run_main
    run_main()


def get_summary_writer(logdir=None, name='test'):
    '''获取 数据写入器
    :param name: 根据名字,自动写入到
    :return:
    '''
    if logdir is None:
        logdir = f'{DEFAULT_PATH}/{name}'
    return SummaryWriter(logdir=logdir)

if __name__ == '__main__':
    run_server()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/6/4 18:13:09

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import inspect, sys, contextlib, types
#===============================================================================
# 
#===============================================================================

__all__ = ['dump_description_help']

@contextlib.contextmanager
def _stdout_file(filename):
    org = sys.stdout
    try:
        with open(filename , 'w') as f:
            sys.stdout = f
            yield
    finally:
        sys.stdout = org


_obj = object()

def dump_description_help(obj, help_file=False, skip_object = True):
    if obj is None:
        print('None')
        return

    if help_file:
        with _stdout_file('help-%s.txt' % _get_name(obj) ):
            help(obj)

    allFunc = []
    allBuiltIn = []
    allAttr = []
    allClass = []
    allModules = []
    allWMethWrapper = []

    print('\n# -----', type(obj))

    for k,v in _get_attr_pair(obj):

        if k.startswith('_') and not k.startswith('__'):
            continue

        if skip_object and hasattr(_obj, k) and k != '__init__':
            continue

        if inspect.isroutine(v):
            if inspect.isbuiltin(v):
                if not k.startswith("_"):
                    allBuiltIn.append((k, v))
            elif isinstance(k, types.MethodWrapperType):
                if k == "__init__":
                    allWMethWrapper.insert(0, (k,v))
                else:
                    allWMethWrapper.append((k,v))
            else:
                allFunc.append((k,v))
        else:
            if inspect.isclass(v):
                allClass.append((k,v))
            elif inspect.ismodule(v):
                allModules.append((k,v))
            else:
                allAttr.append((k, v))

    if allModules:
        print('# module ---->')
        for k,v in allModules:
            print('\t# %s [%s]' %(k, _get_doc(v)) )

    if allClass:
        print('# class ---->')
        for k,v in allClass:
            print('\t# %s [%s]' %(k, _get_doc(v)) )

    if allAttr:
        print('# attr ---->')
        for k,v in allAttr:
            try:
                if k == '__builtins__':
                    print('\t#%s [%s]' %(k, _get_doc(v)) )
                else:
                    print('\t#%s[%s] [%s]' %(k, v, _get_doc(v)) )
            except Exception as e:
                print('\t#%s [%s]' % (k, _get_doc(v)))

    if allWMethWrapper:
        print('# methoid-wrapper ---->')
        for k,v in allWMethWrapper:
            try:
                print('\t#%s[%s] [%s]' %(k, v, _get_doc(v)) )
            except Exception as e:
                print('\t#%s [%s]' % (k, _get_doc(v)))

    if allBuiltIn:
        print('# builtin ---->')
        for k,v in allBuiltIn:
            print('\t# %s [%s]' %(k, _get_doc(v)) )

    if allFunc:
        print('# func ---->')
        for k,v in allFunc:
            print('\t# %s%s [%s]' %(k, _get_signature(v), _get_doc(v)) )

    print()


def _get_doc(v):
    if isinstance(v, str):
        return  '`str`'
    if isinstance(v, bytes):
        return  '`bytes`'
    elif isinstance(v, int):
        return  '`integer`'
    elif isinstance(v, dict):
        return  '`dict`'
    elif isinstance(v, set):
        return  '`set`'
    elif isinstance(v, frozenset):
        return '`fronzenset`'
    elif isinstance(v, float):
        return '`float`'
    elif isinstance(v, tuple):
        return '`tuple`'
    elif isinstance(v, list):
        return '`list`'

    try:
        doc = inspect.getdoc(v)
        return doc.replace('\n', ' ')[0:250]
    except Exception as e:
        return  ''

def _get_signature(v):
    try:
        return str(inspect.signature(v)).replace(', /', '')
    except:
        return "(...)"

def _get_name(v):
    try:
        name = v.__name__
    except Exception:
        try:
            name = v.__class__.__name__
        except Exception:
            name = 'none'
    return name


def _getmembers(object, predicate=None):
    """Return all members of an object as (name, value) pairs sorted by name.
    Optionally, only return members that satisfy a given predicate."""
    if inspect.isclass(object):
        mro = (object,) + inspect.getmro(object)
    else:
        mro = ()

    results = []
    processed = set()
    names = dir(object)

    try:
        for base in object.__bases__:
            for k, v in base.__dict__.items():
                if isinstance(v, types.DynamicClassAttribute):
                    names.append(k)
    except AttributeError:
        pass

    for key in names:
        try:
            value = getattr(object, key)
            # handle the duplicate key
            if key in processed:
                raise AttributeError
        except AttributeError:
            for base in mro:
                if key in base.__dict__:
                    value = base.__dict__[key]
                    break
            else:
                # could be a (currently) missing slot member, or a buggy
                # __dir__; discard and move on
                continue
        except Exception:
            pass

        if not predicate or predicate(value):
            results.append((key, value))
        processed.add(key)
    results.sort(key=lambda pair: pair[0])
    return results

def _get_attr_pair(obj):
    '''
    获取可能的属性列表
    '''
    datalist = []
    if hasattr(obj, '__all__'):
        for k in getattr(obj, '__all__'):
            if hasattr(obj, k):
                datalist.append( (k, getattr(obj, k)) )
        datalist.sort(key=lambda kv: kv[0])
    else:
        datalist = _getmembers(obj)
    return datalist



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/18 00:22:29

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch.optim as optim
#===============================================================================
# 
#===============================================================================

_optim_map = {
    'adam': optim.Adam,
    'SGD' : optim.SGD,
    'Adadelta': optim.Adadelta,
    'RMSprop':optim.RMSprop,
}

def get_optim(model, name, **kws):
    pass

    # if opt.optim == 'Adam':
    #     optimizer = optim.Adam(filter(lambda p: p.requires_grad, model.parameters()),
    #                            lr=opt.learning_rate,
    #                            betas=(opt.optim_alpha, opt.optim_beta),
    #                            eps=opt.optim_epsilon,
    #                            weight_decay=opt.weight_decay)
    # elif opt.optim == 'SGD':
    #     optimizer = optim.SGD(filter(lambda p: p.requires_grad, model.parameters()),
    #                           lr=opt.learning_rate,
    #                           momentum=opt.momentum)
    # elif opt.optim == "momSGD":
    #     optimizer = torch.optim.SGD(filter(lambda p: p.requires_grad, model.parameters()),
    #                                 lr=opt.learning_rate,
    #                                 momentum=opt.momentum)
    # elif opt.optim == 'Adadelta':
    #     optimizer = optim.Adadelta(filter(lambda p: p.requires_grad, model.parameters()),
    #                                lr=opt.learning_rate,
    #                                weight_decay=opt.weight_decay)
    # elif opt.optim == 'RMSprop':
    #     optimizer = optim.RMSprop(filter(lambda p: p.requires_grad, model.parameters()), lr=opt.learning_rate)
    # else:
    #     logging.error("Unknown optimizer: {}".format(opt.optim))
    #     raise Exception("Unknown optimizer: {}".format(opt.optim))
    #
    # # Load the optimizer
    # if opt.resume_from is not None:
    #     optim_path = os.path.join(opt.resume_from, "optimizer.pth")
    #     if os.path.isfile(optim_path):
    #         logging.info("Load optimizer from {}".format(optim_path))
    #         optimizer.load_state_dict(torch.load(optim_path))
    #         opt.learning_rate = optimizer.param_groups[0]['lr']
    #         logging.info("Loaded learning rate is {}".format(opt.learning_rate))
    #
    # return optimizer



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/7/24 11:50:07

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import os
import cv2
import numpy as np
from PIL import Image

from .convert import is_pil_image, pil2cv
#===============================================================================
# 
#===============================================================================

__all__ = ['read_cv_img', 'read_pil_img', 'read_sk_img', 'read_cv_gray', 'read_cv_video', 'write_cv_img']

def read_cv_img(file, flags = cv2.IMREAD_COLOR):
    return cv2.imdecode(np.fromfile(file, dtype=np.uint8), flags)

def read_cv_gray(file):
    return read_cv_img(file, cv2.IMREAD_GRAYSCALE)

def read_sk_img(file):
    import skimage

def read_pil_img(file):
    img = Image.open(file)
    return img

def write_cv_img(img_file, data, params=None):
    '''
    img_file : 保存路径
    img : 图片数据
    '''
    name, ext = os.path.splitext(img_file)
    retval, buf = cv2.imencode(ext, data, params)
    return buf.tofile(img_file)

def read_cv_video(fileOrId=0):
    '''fileOrId: 视频文件或者相机id
       ret: 迭代器,返回图像数据
    '''
    try:
        cap = cv2.VideoCapture(fileOrId)

        while True:
            # Capture frame-by-frame
            ret, frame = cap.read()

            if not ret:
                break

            yield frame

    finally:
        # When everything done, release the capture
        cap.release()



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/6/5 17:49:37

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
from PyQt5.QtWidgets import QApplication, QWidget
import sys, inspect

from .ui import *
#===============================================================================
#
#===============================================================================

def qt_run(widget = None, before_run_func = None, after_run_func = None,  *args, **kwargs):

    try:
        app = QApplication(sys.argv)

        if before_run_func:
            before_run_func()

        win = widget
        if inspect.isclass(widget) or inspect.isfunction(widget):
            win = widget(*args, **kwargs)

        if isinstance(win, QWidget):
            win.show()

        ret = app.exec_()

        if after_run_func:
            after_run_func()

        sys.exit(ret)

    except Exception as e:
        import traceback
        traceback.print_exc()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/17 20:19:45

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import inspect
#===============================================================================
# 实现特定格式数据的打印
#===============================================================================

__all__ = [ 'kprint']

def kprint(d, **kws):
    '''多态打印数据类型
    '''
    _kp_data.do_print(d, **kws)

# 打印配置
class _KPData:
    def __init__(self):
        self._print_funcs = {}

    ###################### 打印相关函数
    def do_print(self, d, **kws):
        t =  type(d)
        func = self._print_funcs[t] if t in self._print_funcs else self._def_print
        msg = func(d, **kws)
        print(msg)

    def register_print(self, name):
        def deco(f):
            self._print_funcs[name] = f
            return f
        return deco

    def unregister_print(self, name):
        if name in self._print_funcs:
            del self._print_funcs[name]

    def _def_print(self, d, **kws):
        msg = f'-----id:{id(d)} type:{type(d)}\n'
        msg += f'{d}'
        return msg

_kp_data = _KPData()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/17 20:14:56

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import numpy as np
from laok.core.kprint import _kp_data, kprint
#===============================================================================
# 
#===============================================================================

@_kp_data.register_print(np.ndarray)
def _print_numpy_arr(d, show_data = True):
    '''打印 numpy array数据
    '''
    msg = str(f'-----id:{id(d)} type:{type(d)}\n'
              f'    flags:\n'
              f'          C_CONTIGUOUS:{d.flags.c_contiguous}\n'
              f'          F_CONTIGUOUS:{d.flags.f_contiguous}\n'
              f'          OWNDATA:{d.flags.owndata}\n'
              f'          WRITEABLE:{d.flags.writeable}\n'
              f'          ALIGNED:{d.flags.aligned}\n'
              f'          WRITEBACKIFCOPY:{d.flags.writebackifcopy}\n'
              f'          UPDATEIFCOPY:{d.flags.updateifcopy}\n'
              f'    shape:{d.shape}\n'
              f'    strides:{d.strides}\n'   
              f'    ndim:{d.ndim}\n'
              f'    size:{d.size}\n'
              f'    itemsize:{d.itemsize}\n'
              f'    nbytes:{d.nbytes}\n'
              f'    dtype:{d.dtype}\n')

    if d.size > 1:
        try:
            max, min = d.max(), d.min()
        except:
            max, min = None, None

        msg += str(f'    max:{max}\n'
                   f'    min:{min}\n')

    if show_data:
        msg += f'{d}'
    return msg



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/8/12 10:06:00

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import importlib, os.path, sys
#===============================================================================
#
#===============================================================================

# 将文件当成模块,然后加载
def dy_load_module(codefile, reload = False):
    '''
    动态加载模块
    '''
    if not isinstance(codefile, str):
        return codefile

    try:
        pthDir = None
        pthDir, pthFile = os.path.split(codefile)
        if pthDir:
            sys.path.insert(0, pthDir)
        fName, fExt = os.path.splitext(pthFile)
        mod = importlib.import_module(fName)
        if reload:
            importlib.reload(mod)
    finally:
        if pthDir:
            sys.path.pop(0)
    return mod


def dy_load_module_attr(modfile, search_attrs = None, reload = False):
    ''' 动态加载 mod 里的属性
    :param modfile: 输入模块名字,可以是内置模块,也可以py文件
    :param search_attrs:  搜索相应的属性,支持 列表/字符串
    :param reload: 是否需要 reload
    :return:
    '''
    if not isinstance(modfile, str):
        return modfile

    try:
        #搜索是否存在路径
        pthDir = None
        if modfile.find('/') != -1 or modfile.find('\\') != -1:
            pthDir, pthFile = os.path.split(modfile)
        else:
            pthFile = modfile

        #插入路径
        if pthDir:
            sys.path.insert(0, pthDir)

        #读取模块名字
        module_str, _ = os.path.splitext(pthFile)

        #动态导入模块
        mod = importlib.import_module(module_str)
        if reload:
            importlib.reload(mod)
        #加载属性名字
        instance = mod
        if isinstance(search_attrs, str):
            # 搜索 属性
            for attr_str in search_attrs.split("."):
                instance = getattr(instance, attr_str)

        elif isinstance(search_attrs, (list, tuple) ):
            # 搜索列表中某一个属性
            for attrs in search_attrs:
                attr_list = attrs.split(".")
                if hasattr(instance, attr_list[0]): #如果搜索到 attr1.attr2.…… 级联模式
                    for attr_str in attr_list:
                        instance = getattr(instance, attr_str)
                    break

        return instance
    finally:
        #删除路径
        if pthDir:
            sys.path.pop(0)


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/27 11:07:29

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch
#===============================================================================
# 
#===============================================================================

def _guess_input(input):
    return input

def save_model(model, filename, input = None):
    '''保存 trace 模型,使用 eval模式固化执行路径;
    :param model: 模型
    :param filename: 文件
    :param input: 测试用例
    :return: trace 模型
    '''
    if isinstance(model, torch.jit.TracedModule):
        traced_model = model
    else:
        try:
            mode = model.training
            model.eval()
            traced_model = torch.jit.trace(model, input)
        finally:
            model.train(mode)

    traced_model.save(filename)
    return traced_model



def load_model(filename):
    ''' 加载 trace 模型
    :param filename: 模型文件
    :return: torch script model
    '''
    return torch.jit.load(filename)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/17 14:20:46

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import pickle
import io
#===============================================================================
# 
#===============================================================================
__all__ = ['load']

def load(data):
    '''
    :param data: 支持从 bytes/filename/fileobj
    :return:
    '''
    if isinstance(data , bytes):
        return pickle.loads(data)
    elif isinstance(data, str):
        with open(data, 'rb') as f:
            return pickle.load(f)
    elif isinstance(data, io.FileIO) :
        return pickle.load(data)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/1/18 19:52:47

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import torch as t
from laok.core.kprint import _kp_data, kprint
#===============================================================================
# 
#===============================================================================
@_kp_data.register_print(t.Tensor)
def _print_torch_arr(d, show_data = True):
    '''打印 torch array数据
    '''
    msg = str(f'-----id:{id(d)} type:{type(d)}\n'
              f'    device:{d.device}\n'
              f'    is_cuda:{d.is_cuda}\n'
              f'    is_leaf:{d.is_leaf}\n'
              f'    is_meta:{d.is_meta}\n'
              f'    is_mkldnn:{d.is_mkldnn}\n'
              f'    is_quantized:{d.is_quantized}\n'
              f'    is_sparse:{d.is_sparse}\n'
              f'    dtype:{d.dtype}\n'
              f'    layout:{d.layout}\n'
              f'    name:{d.name}\n'
              f'    requires_grad:{d.requires_grad}\n'
              f'    ndim:{d.ndim}\n'
              f'    shape:{d.shape}\n')
    if show_data:
        msg += f'{d.data}'
    return msg



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/8/7 14:20:19

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''

#===============================================================================
# 
#===============================================================================

class ChainName(object):
    '''用于实现一个 支持级联调用的 魔法类
    '''
    def __init__(self, prefix, callback, sep = "/"):
        self._prefix = prefix
        self._callback = callback
        self._sep = sep

    def __getattr__(self, item):
        next_prefix = self._prefix + self._sep + item
        return ChainName(next_prefix, self._callback)

    def __call__(self, *kargs, **kwargs):
        self._callback(self._prefix, *kargs, **kwargs)

    def __getitem__(self,item):
        self._prefix = item
        return self

    def get_prefix(self):
        return self._prefix


def index_list(s, sep, escape = None):
    '''获取索引
    :param s:
    :param sep:
    :return:
    '''
    ret = []
    idx = -1
    while 1:
        _i = s.find(sep, idx+1)
        if _i == -1:
            break

        if escape and _i > 0 and s[_i-1] == escape:
            continue

        idx = _i
        ret.append(idx)
    return ret

def index_list_remove_in_range(s, idx_list, range_start, range_end):
    '''删除范围内的索引
    :param s:
    :param idx_list: 索引列表
    :param range_start: 起始符号
    :param range_end: 结束符号
    :return: 剔除了不需要的索引
    '''
    if not idx_list:
        return idx_list
    ret_list = idx_list[:]
    i_start = -1
    i_end = -1

    while 1:
        i_start = s.find(range_start, i_start+1)
        i_end = s.find(range_end, i_end+1)

        if i_start == -1 or i_end == -1:
            break

        for i in range(len(ret_list) - 1, -1, -1):
            if i_start < ret_list[i] < i_end:
                ret_list.pop(i)

    return ret_list


def index_list_to_substrs(s, idx_list):
    '''将索引分割成子串列表
    :param s:
    :param idx_list:
    :return:
    '''
    substrs = []
    lastIdx = -1
    for idx in idx_list:
        substrs.append(s[lastIdx+1:idx])
        lastIdx = idx
    substrs.append(s[lastIdx+1:])
    return substrs



from setuptools import setup, find_packages

setup(
    name='laok',
    version='0.2.11',
    keywords=['_lktest', 'dump_description_help', 'some utils'],
    license='MIT License',
    author='Liu Kuan',
    author_email='1306743659@qq.com',
    description='laok test utils lib',
    packages= find_packages()
)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/7/21 10:33:17

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import re, os, sys
from PyQt5.QtWidgets import QWidget, QMainWindow
from PyQt5.uic import loadUi, compileUi
#===============================================================================
#
#===============================================================================

__all__ = ['DyUiMainWindow', 'DyUiWidget', 'compile_ui_file']

def _init_ui_file(obj):
    if hasattr(obj, 'ui_file'):
        ui_file = getattr(obj, 'ui_file')
    else:
        module_file = sys.modules[obj.__module__].__file__
        ui_file = os.path.splitext(module_file)[0] + ".ui"
    if not os.path.exists(ui_file):
        raise ValueError(f"you need config 'ui_file' in your class of[{obj.__class__}]")
    loadUi(ui_file, obj)  # 初始化Ui
    return ui_file

class DyUiMainWindow(QMainWindow):
    def __new__(cls, *args, **kwargs):
        obj = super(DyUiMainWindow, cls).__new__(cls)   # 构造对象
        QMainWindow.__init__(obj)                       # 调用父类
        _init_ui_file(obj)                              # 初始化ui
        return obj

class DyUiWidget(QWidget):
    def __new__(cls, *args, **kwargs):
        obj = super(DyUiWidget, cls).__new__(cls)   # 构造对象
        QWidget.__init__(obj)                       # 调用父类
        _init_ui_file(obj)                          # 初始化ui
        return obj


def compile_ui_file(ui_file, save_file=None):
    '''
        保存 X_ui.py 文件
    '''
    if save_file is None:
        save_file = os.path.splitext(ui_file)[0] + '_ui.py'
    with open(save_file, 'w', encoding='utf8') as f:
        compileUi(uifile=ui_file, pyfile=f)


# def dy_gen_ui_class(ui_file):
#     '''
#     根据 ui路径或者 _ui路径
#     :return:  返回 ui_class
#     '''
#     py_file = ui_file
#     if not ui_file.endswith('.ui'):
#         ui_file = os.path.splitext(ui_file)[0] + ".ui"
#
#     if uipath.endswith('.ui'):
#         pypath = uipath.replace('.ui', '_ui.py')
#         # 如果存在 ui文件,则编译UI
#         if os.path.exists(uipath):
#             with open(py_file, 'w', encoding='utf8') as f:
#                 uic.compileUi(uifile=uipath, pyfile=f)
#
#     # 开始动态导入 py 模块
#     ui_mod = dy_load_module(pypath)
#     for k in dir(ui_mod):
#         if k.startswith('Ui_'):
#             ui_class = getattr(ui_mod, k)
#             return ui_class
#     else:
#         raise RuntimeError("cannot find class Ui_*")
#
# class StaticUiMainWindow(QMainWindow):
#     def __new__(cls, *args, **kwargs):
#         obj = super(StaticUiMainWindow, cls).__new__(cls)   # 构造对象
#         QMainWindow.__init__(obj)                           # 调用父类
#         _init_ui_file(obj)                                  # 初始化ui
#         return obj



#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2020/7/24 11:59:16

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''
import cv2
import numpy as np
from .convert import is_pil_image, pil2cv
#===============================================================================
# 
#===============================================================================

__all__ = ['show_img', 'wait_escape']


def show_img(img, winName='test', w=0, interpolation=cv2.INTER_LINEAR):
    '''
    :param winName: 窗口名字
    :param img: 图片数据
    :param w: 窗口宽度
    :param interpolation:
    :return:
    '''
    if img is None:
        return

    if is_pil_image(img):
        img = pil2cv(img)

    if img.dtype in [np.bool]: #转换bool 类型
        img = img.astype(np.uint8) * 255

    elif img.dtype not in [np.uint8, ]:  #转换 其它类型
        maxV = img.max()
        if 0 < maxV <= 1.0:  #如果是浮点类型
            img = img * 255

        img = img.astype(np.uint8)

    # 将尺寸放缩
    if w != 0:
        _h, _w = img.shape[:2]

        if w == -1:
            w = 1200 if _w > 1200 else _w

        h = 1.0 * _h * w / _w

        showImg = cv2.resize(img, dsize=(int(w), int(h)), interpolation=interpolation)
    else:
        showImg = img

    # 显示
    cv2.imshow(winName, showImg)


def wait_escape(delay=0, closeWindow=False):
    ''' delay: 延时 (ms)
        closeWindow:是否关闭窗口
    '''
    try:
        key = cv2.waitKey(delay)
        if key == 27:#退出
            cv2.destroyAllWindows()
            raise SystemExit
        return key
    finally:
        if closeWindow:
            cv2.destroyAllWindows()


#! /usr/bin/env python
# -*- coding: utf-8 -*-
'''
Created on 2021/3/31 15:46:44

@author: LiuKuan
@copyright: Apache License, Version 2.0
'''

#===============================================================================
# 
#===============================================================================

def file_read_lines(fname, encoding=None, skip_empty = False):
    retList = []
    with open(fname, 'r', encoding=None) as f:
        for line in f:
            line = line.strip()
            if skip_empty and not line:
                continue
            retList.append(line)
    return retList

def file_write_lines(fname, lines, encoding=None):
    with open(fname, 'w', encoding=encoding) as f:
        for line in lines:
            f.write(line)
            f.write('\n')
